# Ajuste no lineal de elipses
# Del paper:
# Sung Joon Ahn, Wolfgang Rauh, Hans-Jürgen Warnecke,
# Least-squares orthogonal distances fitting of circle, sphere,
# ellipse, hyperbola, and parabola
# Pattern Recognition,
# Volume 34, Issue 12, December 2001, Pages 2283-2303
# https://doi.org/10.1016/S0031-3203(00)00152-7
#
# Código en python
# usando Levenberg-Marquardt
# Fraga 11.08.2020
#
import pprime
import numpy as np
import scipy.optimize as lm
import math
import sys

n = len( sys.argv )
if n != 7:
	print( "Args: archivo_puntos a0 b0 Xc0 Yc0 alfa0" )
	sys.exit(1)

# La elipse inicial
va = np.zeros( (5,) )
va[0] = float(sys.argv[2])
va[1] = float(sys.argv[3])
va[2] = float(sys.argv[4])
va[3] = float(sys.argv[5])
va[4] = float(sys.argv[6])

# Leemos los puntos de entrada
P = np.genfromtxt( sys.argv[1], delimiter=' ' )
# print( P.size )
# print( P.shape )
[n, m] = P.shape

# Para el LM se deben calcular los residuos
#   1. Los puntos iniciales están el sistema de coordenadas XY
#   2. Los pasamos al sistema de coordenadas xy
#       p = R(P - Pc)
#   3. Calculamos aqui los puntos p' (sobre la elipse)
#   4. Y regresamos p' al sistema de coordenadas XY
#       P = R^T p + Pc

def residuos( vx ) :
	a = vx[0]
	b = vx[1]
	c = math.cos( vx[4] )
	s = math.sin( vx[4] )

	vr = np.zeros( (2*n, ) )

	j = 0
	while j < n :
		x = P[j][0] - vx[2]
		y = P[j][1] - vx[3]

		xi = c*x + s*y
		yi =-s*x + c*y
		# print( xi, yi )

		# 3 :
		[xp, yp] = pprime.pprime( vx, [xi, yi] )

		# 4 :
		Xp = c*xp - s*yp + vx[2]
		Yp = s*xp + c*yp + vx[3]

		vr[2*j] = P[j][0] - Xp
		vr[2*j+1] = P[j][1] - Yp

		j += 1

	return vr


resultado = lm.least_squares( residuos, va, method='trf' )
# resultado = lm.least_squares( residuos, va, method='lm' )
print( resultado )
